"""Network interface group widget."""

from __future__ import annotations

from textual import on
from textual.app import ComposeResult
from textual.containers import Grid
from textual.events import DescendantFocus, Key
from textual.message import Message
from textual.reactive import var

from flux_networking_shared.tui.messages import FocusTabsRequested
from flux_networking_shared.tui.widgets.interface import Interface


class FocusButtonsRequested(Message):
    """Request focus on buttons container."""


class InterfaceGroup(Grid):
    """Grid container for network interface widgets."""

    class InterfaceModalRequested(Message):
        """Request to open interface configuration modal."""

        def __init__(self, interface: Interface, in_use_vlans: list[int]) -> None:
            super().__init__()
            self.interface = interface
            self.in_use_vlans = in_use_vlans

    class NoInterfacesModalRequested(Message):
        """Request to show 'no interfaces' dialog."""

    focused_child: var[int | str] = var("")

    DEFAULT_CSS = """
    InterfaceGroup {
        grid-size: 2;
        grid-gutter: 1;
    }
    """

    def __init__(self, interfaces: list[Interface] | None = None):
        """Initialize the interface group.

        Args:
            interfaces: List of Interface widgets to display
        """
        super().__init__()
        self.interfaces = interfaces or []

    @property
    def last_interface(self) -> bool:
        """Check if focused on last interface."""
        if isinstance(self.focused_child, str):
            return False
        return self.focused_child + 1 == len(self.children)

    @property
    def first_interface(self) -> bool:
        """Check if focused on first interface."""
        if isinstance(self.focused_child, str):
            return True
        return self.focused_child == 0

    def get_child_interface_names(self, interface: Interface) -> list[str]:
        """Get names of child VLAN interfaces.

        Args:
            interface: Parent interface

        Returns:
            List of child interface names
        """
        return [
            x.interface_name
            for x in self.interfaces
            if x.interface_type == "vlan" and x.parent_ifindex == interface.ifindex
        ]

    def update_from_dict(self, interface_data: dict) -> None:
        """Update an interface from dictionary data.

        Args:
            interface_data: Dictionary with interface data
        """
        name = interface_data.get("name")
        if not name:
            return

        interface = next(
            (x for x in self.interfaces if x.interface_name == name),
            None,
        )

        if not interface:
            # Create new interface
            interface = Interface.from_dict(interface_data)
            interface.id = f"int-{len(self.interfaces)}"
            self.interfaces.append(interface)

            if self.is_mounted:
                self.mount(interface)

            return

        # Update existing interface
        interface.update_from_dict(interface_data)

    def id_to_interface(self, id: int | str) -> Interface:
        """Get interface by numeric ID.

        Args:
            id: Interface ID number

        Returns:
            Interface widget
        """
        child: Interface = self.get_child_by_id(f"int-{id}")  # type: ignore
        return child

    def compose(self) -> ComposeResult:
        """Compose the interface grid."""
        yield from self.interfaces

    def in_use_vlans(self, interface: Interface) -> list[int]:
        """Get VLAN IDs already in use by this interface.

        Args:
            interface: Parent interface

        Returns:
            List of VLAN IDs
        """
        return [
            x.vlan
            for x in self.interfaces
            if x.parent_ifindex == interface.ifindex and x.vlan is not None
        ]

    async def add_vlan_subinterface(
        self, parent: Interface, vlan_id: int, set_loading: bool = False
    ) -> Interface:
        """Add a new VLAN subinterface.

        Args:
            parent: Parent interface
            vlan_id: VLAN ID
            set_loading: Whether to show loading indicator

        Returns:
            The new VLAN interface widget
        """
        vlan_interface = Interface(
            type="vlan",
            name=f"{parent.interface_name}.{vlan_id}",
            state="DOWN",
            ifindex=-1,
            mac_addr=parent.mac_addr,
            ip_allocation="disabled",
            network_vendor="",
            network_type="",
            vlan=vlan_id,
            parent_ifindex=parent.ifindex,
            id=f"int-{len(self.interfaces)}",
        )
        vlan_interface.loading = set_loading

        self.interfaces.append(vlan_interface)

        index = len(self.interfaces) - 1

        await self.mount(vlan_interface)
        self.set_child_focus(first=False, index=index)

        return vlan_interface

    def set_child_focus(self, first: bool = True, index: int | None = None) -> None:
        """Set focus to a child interface.

        Args:
            first: If True, focus first interface
            index: Specific index to focus
        """
        if first:
            idx = 0
        elif index is not None:
            idx = index
        else:
            idx = len(self.children) - 1

        self.focused_child = idx

    def watch_focused_child(self, id: int | str) -> None:
        """Focus the child when focused_child changes."""
        if id == "":
            return

        interface = self.id_to_interface(id)
        interface.focus()

    def blur_children(self) -> None:
        """Blur all children."""
        self.focused_child = ""

    def on_descendant_focus(self, event: DescendantFocus) -> None:
        """Handle descendant focus events."""
        if self.focused_child == "" and event.widget.id:
            try:
                self.focused_child = int(event.widget.id.split("-")[-1])
            except (ValueError, IndexError):
                pass

    @on(Interface.InterfaceClicked)
    def on_interface_clicked(self, event: Interface.InterfaceClicked) -> None:
        """Handle interface click."""
        self.post_message(
            self.InterfaceModalRequested(
                event.interface, self.in_use_vlans(event.interface)
            )
        )

    def on_key(self, event: Key) -> None:
        """Handle keyboard navigation."""
        key = event.name

        if key not in [
            "left",
            "up",
            "right",
            "down",
            "tab",
            "shift_tab",
            "enter",
            "space",
        ]:
            return

        if (
            (not self.last_interface and key == "tab")
            or (not self.first_interface and key == "shift_tab")
            or key == "down"
            or key == "up"
        ):
            event.stop()

        interface_count = len(self.children)
        if interface_count == 0:
            return

        row_count = (interface_count // 2) + (interface_count % 2)
        current_focused = (
            self.focused_child if isinstance(self.focused_child, int) else 0
        )
        current_row = current_focused // 2

        if key == "tab":
            if self.last_interface:
                self.focused_child = ""
                self.post_message(FocusButtonsRequested())
            else:
                self.focused_child = current_focused + 1
        elif key == "shift_tab":
            if self.first_interface:
                self.focused_child = ""
                self.post_message(FocusTabsRequested())
            else:
                self.focused_child = current_focused - 1
        elif key == "right":
            self.focused_child = (current_focused + 1) % interface_count
        elif key == "left":
            self.focused_child = (
                current_focused + interface_count - 1
            ) % interface_count
        elif key == "up":
            if current_row == 0:
                self.focused_child = ""
                self.post_message(FocusTabsRequested())
            else:
                self.focused_child = current_focused - 2
        elif key == "down":
            if row_count > current_row + 1:
                self.focused_child = min(interface_count - 1, current_focused + 2)
            else:
                self.focused_child = ""
                self.post_message(FocusButtonsRequested())

        elif key in ["enter", "space"]:
            interface = self.id_to_interface(current_focused)
            self.post_message(
                self.InterfaceModalRequested(interface, self.in_use_vlans(interface))
            )



"""Network summary widget."""

from __future__ import annotations

from textual import on
from textual.app import ComposeResult
from textual.events import Key
from textual.widget import Widget

from flux_networking_shared.tui.messages import FocusTabsRequested
from flux_networking_shared.tui.widgets.interface_group import FocusButtonsRequested
from flux_networking_shared.tui.widgets.network_overview import NetworkOverview


class NetworkSummary(Widget, can_focus=True):
    """Container widget that displays network overview and connectivity status."""

    BORDER_TITLE = "Network Summary"

    DEFAULT_CSS = """
        NetworkSummary {
            height: auto;
            border: solid $primary;
            padding: 1;
        }

        NetworkSummary.--highlight {
            border: double $primary;
            background: $panel;
        }
    """

    def __init__(self) -> None:
        super().__init__()
        self.network_overview = NetworkOverview()

    def compose(self) -> ComposeResult:
        yield self.network_overview

    def on_key(self, event: Key) -> None:
        """Handle up/down navigation."""
        event_name = event.name

        if event_name not in ["up", "down"]:
            return

        event.stop()

        msg_cls = FocusTabsRequested if event_name == "up" else FocusButtonsRequested

        self.post_message(msg_cls())

    def set_network_address(self, address: str) -> None:
        """Set the primary network address."""
        self.network_overview.network_address = address

    def set_default_route(self, route: str) -> None:
        """Set the default route."""
        self.network_overview.default_route = route

    def set_dns(self, dns: str) -> None:
        """Set DNS server info."""
        self.network_overview.dns = dns

    def set_upnp_details(self, enabled: str, port: str) -> None:
        """Set UPnP status details."""
        self.network_overview.upnp_enabled = enabled
        self.network_overview.upnp_port = port

    def set_connectivity(self, state: str) -> None:
        """Set connectivity status."""
        self.network_overview.connectivity = state

    def set_name_resolution(self, state: str) -> None:
        """Set name resolution status."""
        self.network_overview.name_resolution = state

    def watch_has_focus(self, has_focus: bool) -> None:
        """Toggle highlight class on focus."""
        super().watch_has_focus(has_focus)
        self.set_class(has_focus, "--highlight")

    @on(NetworkOverview.NetworkAvailableMessage)
    def on_network_available(
        self, event: NetworkOverview.NetworkAvailableMessage
    ) -> None:
        """Handle network availability changes.

        When network becomes available, the parent screen may want to
        initiate connectivity tests to external services.
        """
        # Parent screen should handle this event to start/stop
        # connectivity monitoring as needed
        pass



"""Label and switch combo widget."""

from __future__ import annotations

from textual.app import ComposeResult
from textual.containers import Horizontal
from textual.dom import NoMatches
from textual.reactive import var
from textual.widgets import Label, Switch


class LabelSwitch(Horizontal):
    """Horizontal container with a label and toggle switch."""

    DEFAULT_CSS = """
        LabelSwitch {
            height: auto;
            width: auto;

            & > Label {
                height: 3;
                content-align: center middle;
            }
        }
    """

    label = var("")
    value = var(False)

    def __init__(
        self, label: str, *, value: bool = True, id: str | None = None
    ) -> None:
        super().__init__(id=id)
        self._label = label
        self._value = value

    def compose(self) -> ComposeResult:
        yield Label(self._label)
        yield Switch(self._value)

    def on_mount(self) -> None:
        self.label = self._label
        self.value = self._value

    def on_switch_changed(self, event: Switch.Changed) -> None:
        """Sync reactive value when switch changes."""
        self.value = event.value

    def watch_label(self, old: str, new: str) -> None:
        if old == new:
            return

        self._label = new

        try:
            label = self.query_one(Label)
        except NoMatches:
            return

        label.update(new)

    def watch_value(self, old: bool, new: bool) -> None:
        if old == new:
            return

        self._value = new

        try:
            switch = self.query_one(Switch)
        except NoMatches:
            return

        switch.value = new



from __future__ import annotations

import asyncio
from typing import Any
from itertools import count
from json import JSONDecodeError

from socket import AF_INET

from yarl import URL

from aiohttp import (
    ClientConnectorError,
    ClientOSError,
    ClientSession,
    ClientTimeout,
    ConnectionTimeoutError,
    ContentTypeError,
    TCPConnector,
    ClientResponse,
    BasicAuth,
)

import pwd
import os

from flux_networking_shared.log import log


class ExecBinaryError(ChildProcessError):
    def __init__(self, cmd: list[str], stdout: bytes, stderr: bytes):
        self.__cmd = cmd
        self.__stderr = stderr
        super().__init__([cmd, (stdout, stderr)])

    def stderr(self):
        return self.__stderr.decode(errors="ignore")

    def __str__(self):
        return "Failed to execute '{}', error: {}".format(
            " ".join(self.__cmd), self.stderr()
        )


def merge_dict(a: dict, b: dict, path=[]):
    """Recursively merge dicts (opinionated)

    Args:
        a (dict): The mutated dict dict
        b (dict): The overriding dict
        path (list, optional): where in the chain we are at. Defaults to [].

    Returns:
        _type_: The merged dict. dict b values take precedence.
    """
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge_dict(a[key], b[key], path + [str(key)])
            else:
                a[key] = b[key]
        else:
            a[key] = b[key]
    return a


def demote(user_uid, user_gid):
    def result():
        os.setgid(user_gid)
        os.setuid(user_uid)

    return result


async def exec_binary(
    cmd: list[str],
    expect_returncode: int = 0,
    cwd: str | None = None,
    user: str | None = None,
    env: dict | None = None,
) -> str:
    if len(cmd) == 0:
        raise ChildProcessError("Cannot execute empty cmd")

    extra_params: dict[str, Any] = {}

    if cwd is not None:
        extra_params["cwd"] = cwd

    if user is not None:
        try:
            pw_record = pwd.getpwnam(user)
        except KeyError:
            raise ChildProcessError("User not found")

        extra_params["preexec_fn"] = demote(pw_record.pw_uid, pw_record.pw_gid)

    if env:
        extra_params["env"] = env

    try:
        proc = await asyncio.create_subprocess_exec(
            *cmd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            **extra_params,
        )
    except FileNotFoundError:
        raise ChildProcessError(f"Binary not found: {cmd[0]}, is it installed?")

    try:
        stdout, stderr = await proc.communicate()
        if proc.returncode != expect_returncode:
            raise ExecBinaryError(cmd, stdout, stderr)
    except asyncio.CancelledError:
        if proc.returncode is None:
            proc.terminate()
            return_code = await proc.wait()
            log.info(
                "Cmd: %s cancelled and terminated. Exit code: %s",
                cmd[0],
                return_code,
            )
        raise asyncio.CancelledError()

    return stdout.decode()


async def do_http(
    url: URL | str,
    verb: str = "get",
    *,
    data: Any = None,
    connect_timeout: int = 3,
    max_tries: int = 3,
    retry_interval: int = 1,
    credentials: dict | None = None,
    headers: dict | None = None,
    verify_ssl: bool | None = None,
    total_timeout: int | None = None,
) -> Any:
    timeout = ClientTimeout(connect=connect_timeout, total=total_timeout)
    conn = TCPConnector(family=AF_INET)

    auth = None
    res = None

    if credentials:
        # user password
        auth = BasicAuth(*credentials)

    async with ClientSession(
        connector=conn,
    ) as session:
        for attempt in count(start=1):
            if max_tries and attempt > max_tries:
                break

            remaining = max(0, max_tries - attempt)

            try:
                method = getattr(session, verb)
                async with method(
                    url,
                    timeout=timeout,
                    auth=auth,
                    headers=headers,
                    json=data,
                    ssl=verify_ssl,
                ) as resp:
                    resp: ClientResponse  # type: ignore

                    if resp.status in [429, 500, 502, 503, 504]:
                        log.debug("bad response: %s", resp.status)
                        if remaining or not max_tries:
                            await asyncio.sleep(retry_interval)

                        continue

                    if resp.status in [401, 403, 404]:
                        log.info("URL: %s, status: %s", url, resp.status)
                        return None

                    try:
                        res = await resp.json()
                    except (JSONDecodeError, ContentTypeError):
                        res = None
                    break

            except (
                ClientConnectorError,
                ConnectionTimeoutError,
                ClientOSError,
                asyncio.TimeoutError,
            ):
                # ClientOSError: have seen Connection reset by peer
                log.info(
                    "Unable to connect to: %s. Retries: %s remaining. Interval: %ss",
                    url,
                    remaining,
                    retry_interval,
                )

                if remaining or not max_tries:
                    await asyncio.sleep(retry_interval)

    return res



from ast import literal_eval
from configparser import ConfigParser


class ConfigParserDict(dict):
    def items(self):
        for k, v in super().items():
            if v.startswith("[") and v.endswith("]"):
                # change this to json parse
                for i in literal_eval(v):
                    yield k, i
            else:
                yield k, v


class SystemdConfigParser(ConfigParser):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._dict = ConfigParserDict

    def optionxform(self, optionstr: str) -> str:
        # this stops the keys being lowercased
        return optionstr



"""Confirmation dialog screen."""

from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal, Vertical
from textual.screen import ModalScreen
from textual.widgets import Button, Label


class ConfirmExitScreen(ModalScreen[bool]):
    """Modal screen for confirming destructive actions."""

    DEFAULT_CSS = """
        ConfirmExitScreen {
            align: center middle;
        }

        ConfirmExitScreen > Vertical {
            width: 50%;
            height: auto;
            border: solid $primary;
            align: center middle;
        }

        ConfirmExitScreen > Vertical > Label {
            height: 5;
            width: 1fr;
            column-span: 2;
            content-align: center middle;
            text-align: center;
            color: $error;
        }

        ConfirmExitScreen > Vertical > Horizontal {
            height: auto;
            align: center middle;
        }

        ConfirmExitScreen > Vertical > Horizontal > Button {
            margin: 0 1 1 1;
        }
    """

    BINDINGS = [Binding("escape", "dismiss", "Dismiss", show=False)]

    def __init__(self, message: str, action_label: str) -> None:
        """Initialize the confirmation screen.

        Args:
            message: The confirmation message to display
            action_label: Label for the action button (e.g., "Delete")
        """
        super().__init__()
        self.message = message
        self.action_label = action_label

    def compose(self) -> ComposeResult:
        with Vertical():
            yield Label(self.message, id="confirm-exit")
            with Horizontal(classes="button-container"):
                yield Button("Back", id="cancel")
                yield Button(self.action_label, id="action", variant="error")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        """Handle button press."""
        action = event.button.id == "action"
        self.dismiss(action)



"""Network TUI screens."""

from flux_networking_shared.tui.screens.confirm_exit import ConfirmExitScreen
from flux_networking_shared.tui.screens.interface_modal_screen import (
    InterfaceModalScreen,
)
from flux_networking_shared.tui.screens.network_screen import NetworkScreen

__all__ = [
    "ConfirmExitScreen",
    "InterfaceModalScreen",
    "NetworkScreen",
]



# import logging

# logging.basicConfig()
# logging.getLogger().setLevel(logging.DEBUG)

import asyncio
import os
import subprocess
from collections.abc import Sequence
from typing import Callable

from flux_networking_shared.models.network import NetworkInterface

# just use all the subiquity / probert models
from probert.network import Link, NetworkEventReceiver, UdevObserver

# Standard interface flags (net/if.h)
IFF_UP = 0x1  # Interface is up.
IFF_BROADCAST = 0x2  # Broadcast address valid.
IFF_DEBUG = 0x4  # Turn on debugging.
IFF_LOOPBACK = 0x8  # Is a loopback net.
IFF_POINTOPOINT = 0x10  # Interface is point-to-point link.
IFF_NOTRAILERS = 0x20  # Avoid use of trailers.
IFF_RUNNING = 0x40  # Resources allocated.
IFF_NOARP = 0x80  # No address resolution protocol.
IFF_PROMISC = 0x100  # Receive all packets.
IFF_ALLMULTI = 0x200  # Receive all multicast packets.
IFF_MASTER = 0x400  # Master of a load balancer.
IFF_SLAVE = 0x800  # Slave of a load balancer.
IFF_MULTICAST = 0x1000  # Supports multicast.
IFF_PORTSEL = 0x2000  # Can set media type.
IFF_AUTOMEDIA = 0x4000  # Auto media select active.

IFA_F_PERMANENT = 0x80

NETDEV_IGNORED_IFACE_TYPES = [
    "lo",
    "bridge",
    "tun",
    "tap",
    "dummy",
    "sit",
    "can",
    "???",
]


def _clean_env(env, *, locale=True):
    if env is None:
        env = os.environ.copy()
    else:
        env = env.copy()
    if locale:
        env["LC_ALL"] = "C"

    return env


def run_command(
    cmd: Sequence[str],
    *,
    input=None,
    stdout=subprocess.PIPE,
    stderr=subprocess.PIPE,
    encoding="utf-8",
    errors="replace",
    env=None,
    clean_locale=True,
    **kw,
) -> subprocess.CompletedProcess:
    """A wrapper around subprocess.run with logging and different defaults.

    We never ever want a subprocess to inherit our file descriptors!
    """
    if input is None:
        kw["stdin"] = subprocess.DEVNULL
    else:
        input = input.encode(encoding)
    print(f"run_command called: {cmd}")
    try:
        cp: subprocess.CompletedProcess = subprocess.run(
            cmd,
            input=input,
            stdout=stdout,
            stderr=stderr,
            env=_clean_env(env, locale=clean_locale),
            **kw,
        )
        if encoding:
            if isinstance(cp.stdout, bytes):
                cp.stdout = cp.stdout.decode(encoding)
            if isinstance(cp.stderr, bytes):
                cp.stderr = cp.stderr.decode(encoding)
    except subprocess.CalledProcessError as e:
        print(f"run_command: {e}")
        raise
    else:
        print(f"run_command: {cp.args} exited with code: {cp.returncode}")
        return cp


class Receiver(NetworkEventReceiver):
    def __init__(self, callback: Callable[[NetworkInterface], None]):
        self.all_links: dict[int, Link] = {}
        self.callback = callback
        # self.route_data = []

    def new_link(self, ifindex: int, link: Link):
        if link.type in NETDEV_IGNORED_IFACE_TYPES:
            return

        self.all_links[ifindex] = link

        # we assume all new links are down
        # is_connected = False

        flags = getattr(link, "flags", 0)
        is_connected = bool(flags & IFF_UP)

        print(f"Observer new link: {is_connected} for link: {link.name}")

        # I was doing this only on update link, but if the link is actually down
        # you don't get the update
        self.callback(
            NetworkInterface.from_probe_data(link.serialize(), is_connected),
        )

    def route_change(self, action: str, data):
        # self.route_data.append(data)
        pass

    def update_link(self, ifindex: int):
        link = self.all_links.get(ifindex)

        if not link:
            return

        flags = getattr(link, "flags", 0)
        is_connected = bool(flags & IFF_UP)

        print(f"Observer updated link: {is_connected} for link: {link.name}")

        self.callback(
            NetworkInterface.from_probe_data(link.serialize(), is_connected),
        )

    def del_link(self, ifindex):
        try:
            del self.all_links[ifindex]
        except KeyError:
            pass


class Observer:
    def __init__(self, receiver: Receiver) -> None:
        self.receiver: Receiver | None = receiver
        self.observer: UdevObserver | None = UdevObserver(self.receiver)
        self._observer_fds: list[int] = []

        self._watching = False

    def start(self) -> None:
        self._observer_fds.extend(self.observer.start())
        self.start_watching()

    def stop(self) -> None:
        self.stop_watching()

        self._observer_fds = []
        self.observer = None
        self.receiver = None

    def stop_watching(self):
        if not self._watching:
            return

        loop = asyncio.get_running_loop()

        for fd in self._observer_fds:
            loop.remove_reader(fd)

        self._watching = False

    def start_watching(self):
        if self._watching:
            return

        self._watching = True

        loop = asyncio.get_running_loop()

        for fd in self._observer_fds:
            loop.add_reader(fd, self._data_ready, fd)

    def _data_ready(self, fd):
        cp = run_command(["udevadm", "settle", "-t", "0"])
        if cp.returncode != 0:
            print("waiting 0.1 to let udev event queue settle")
            self.stop_watching()
            loop = asyncio.get_running_loop()
            loop.call_later(0.1, self.start_watching)
            return
        self.observer.data_ready(fd)


if __name__ == "__main__":

    async def main():
        obs = Observer()
        obs.start()

    loop = asyncio.new_event_loop()
    loop.create_task(main())
    loop.run_forever()
